/* WinPT.cpp - Windows Privacy Tray (WinPT)
 *	Copyright (C) 2000-2005 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptErrors.h"
#include "wptGPG.h"
#include "wptRegistry.h"
#include "wptCommonCtl.h"
#include "wptDlgs.h"
#include "wptNLS.h"
#include "wptKeyserver.h"
#include "wptCard.h"
#include "wptFileManager.h"
#include "wptContext.h"
#include "wptCardEdit.h"
#include "wptCrypto.h"

#define MIN_GPG_VER   "1.4.2"	/* Minimal GPG version. */
#define MIN_GPGME_VER "1.2.0"	/* Minimal GPGME version. */


HINSTANCE glob_hinst;	/* global instance for the dialogs */
HWND glob_hwnd;		/* global window handle for the dialogs */
HWND activ_hwnd;
LOCK mo_file;
int scard_support = 0;
int debug = 0;
int mobile = 0;
int gpg_read_only = 0;
char gpgver[3];


/* Load the key cache and rebuild the signature cache. */
static void
update_keycache (HWND hwnd)
{
    refresh_cache_s rcs = {0};
    rcs.kr_reload = 0; 
    rcs.kr_update = 1;
    rcs.tr_update = 1;
    DialogBoxParam (glob_hinst, (LPCSTR)IDD_WINPT_KEYCACHE, hwnd,
		    keycache_dlg_proc, (LPARAM)&rcs);
}


/* Set GPGME debug mode. If @val is 0, the debug mode is disabled. */
void
gpg_set_debug_mode (int val)
{	
    if (val)
	putenv ("GPGME_DEBUG=5:gpgme.dbg");
    else
	putenv ("GPGME_DEBUG=");
}


/* Return the name of the gettext language file. */
static char*
get_gettext_lang (void)
{    
    char *fname;
    fname = get_reg_entry_mo ();
    if (!fname)
	return NULL;
    return fname;
}


/* Initialize the gettext sub system. */
static void
load_gettext (int prev_inst)
{
    char *nls = NULL;
    char *file = NULL;

    nls = get_gettext_lang ();
    if (nls) {
	set_gettext_file ("winpt", nls);
	file = make_filename (nls, "winpt", "mo");
	if (!file_exist_check (nls) && init_file_lock (&mo_file, file))  {
	    if (!prev_inst)
		msg_box (NULL, _("Could not initizalize file lock.\n"
				 "Native Language Support"),
			 _("WinPT Error"), MB_ERR);
	}
	free_if_alloc (nls);
	free_if_alloc (file);
    }
}


/* Load the GPG environment. On the first start, some
   checks are performed to find out in what state GPG is.
   Return value: 0  everything OK.
                 >0  fatal error.
		 -1 public keyring is empty or does not exist. */
static int
load_gpg_env (void)
{
    SECURITY_ATTRIBUTES sec_attr;
    char *p;
    char *pkr;

    p = get_reg_entry_gpg4win ("gpg.exe");
    if (!p)
	return (1);
    if (file_exist_check (p)) {
	free_if_alloc (p);
	return (1);
    }
    free_if_alloc (p);
    p = multi_gnupg_path (0);
    if (p && dir_exist_check (p)) {
	memset (&sec_attr, 0, sizeof (sec_attr));
	sec_attr.nLength = sizeof (sec_attr);
	if (!CreateDirectory (p, &sec_attr)) {
	    msg_box (NULL, _("Could not create GPG home directory"), 
		     _("WinPT Error"), MB_ERR);
	    free_if_alloc (p);
	    return (2);
	}
    }
    pkr = make_filename (p, "pubring", "gpg");
    free_if_alloc (p);
    if (!pkr)
	return -1;
    if (get_file_size (pkr) == 0) {
	free_if_alloc (pkr);
	return -1;
    }
    return 0;
}

/* check if the default key from the gpg.conf file is available in the
   keyring. if not, bail out because encryption won't work properly then. */
static int
check_default_key (gpg_keycache_t kc)
{
    gpgme_key_t key;
    gpgme_error_t err = GPG_ERR_NO_ERROR;
    char *defkey;

    defkey = get_gnupg_default_key ();
    if (defkey)
	err = gpg_keycache_find_key (kc, defkey, 0, &key);
    else
	msg_box (NULL, _("No useable secret key found."), 
		 _("WinPT Error"), MB_ERR);
    free_if_alloc (defkey);
    return err? -1 : 0;
}


/* Return the WinPT program file name (with full pathname). */
static const char*
get_prog_part (const char * fname, int use_cwd)
{
    static char program[512];
    char currdir[256];
    char *cmd = NULL;
    int j;
	
    memset (currdir, 0, DIM (currdir));
    memset (program, 0, DIM (program));
	
    if (use_cwd) {
	GetCurrentDirectory (DIM (currdir)-1, currdir);
	_snprintf (program, DIM (program)-1, "%s\\%s", currdir, fname);
    } 
    else {
	cmd = GetCommandLine ();
	if (cmd == NULL)
	    return NULL;
	strncpy (currdir, cmd, sizeof (currdir)-1);
	j = strlen (currdir);
	while (j--) {
	    if (currdir[j] == '\\')
		break;
	}
	currdir[j] = 0;
	_snprintf (program, DIM (program)-1, "%s\\%s", currdir + 1, fname);
    }
    return program;
}


/* Check that the underlying crypto engine fullfills the minimal
   requirements so all commands work properly. */
static bool
check_crypto_engine (void)
{
    int ma=1, mi=4, pa=2; /* GPG 1.4.2 */
    int rc;

    rc = check_gnupg_engine (&ma, &mi, &pa);
    if (rc == -1) {
	msg_box (NULL, _("Could not read GnuPG version."), 
		 _("WinPT Error"), MB_ERR);
	return false;
    }
    else if (rc) {
	log_box (_("WinPT Error"), MB_ERR,
		 _("Sorry, you need a newer GPG version.\n"
		   "GPG version %d.%d.%d required GPG version "MIN_GPG_VER),
		   ma, mi, pa);
	return false;
    }
    /* We enable smartcard support for GPG: >= 2 or >= 1.4.3 */
    if (ma > 1 || pa >= 3)    
	scard_support = 1;

    gpgver[0] = ma;
    gpgver[1] = mi;
    gpgver[2] = pa;
    return true;
}


/* Try to load the keyserver config file. If @quiet is 1
   do not show any errors. */
static int
load_keyserver_conf (int quiet)
{
    const char *t;
    int rc;

    if (reg_prefs.kserv_conf)
	t = reg_prefs.kserv_conf;
    else if (!file_exist_check (get_prog_part ("keyserver.conf", 0)))
	t = get_prog_part ("keyserver.conf", 0);
    else 
	t = "keyserver.conf";
    rc = kserver_load_conf (t);
    if (rc && !quiet)
	msg_box (NULL, winpt_strerror (rc), _("Keyserver"), MB_ERR);
    return rc;
}


/* Enable the mobility mode. */
static void
enable_mobile_mode (void)
{
    memset (&reg_prefs, 0, sizeof (reg_prefs));
    reg_prefs.always_trust = 0;
    reg_prefs.auto_backup = 0;
    reg_prefs.cache_time = 0;
    reg_prefs.expert = 0;
    reg_prefs.keylist_mode = 1;
    reg_prefs.kserv_conf = m_strdup ("keyserver.conf");
    reg_prefs.no_zip_mmedia = 1;
    reg_prefs.use_tmpfiles = 1;
    reg_prefs.word_wrap = 80;
    reg_prefs.use_viewer = 0; /* XXX */
}


/* Main entry point. */
int WINAPI
WinMain (HINSTANCE hinst, HINSTANCE hprev, LPSTR cmdline, int showcmd)
{
    WNDCLASS wc = {0, winpt_main_proc, 0, 0, hinst, 0, 0, 0, 0, PGM_NAME};
    HACCEL accel_tab;
    MSG msg;
    HWND hwnd = NULL;
    WORD ver[3], ptdver[4];
    int rc, ec, created = 0;
    int first_start = 0, start_gpgprefs = 0;
    int winpt_inst_found = 0;
    int start_manager = 0;
    const char *s;

    glob_hinst = hinst;
    if (cmdline && stristr (cmdline, "--stop")) {
	hwnd = FindWindow ("WinPT", "WinPT");
	if (hwnd != NULL)
	    PostMessage (hwnd, WM_DESTROY, 0, 0);
	return 0;
    }

    /*
    OSVERSIONINFO osinf;
    memset (&osinf, 0, sizeof (osinf));
    if (GetVersionEx (&osinf) && 
	osinf.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS &&
	osinf.dwMinorVersion == 0) {
	msg_box (NULL, "WinPT propably does not work on Windows 95 without restrictions",
	         "WinPT Warning", MB_INFO);
    }
    */

    #ifdef _DEBUG
    gpg_set_debug_mode (1);
    debug = 1;
    #endif

    get_file_version ("WinPT.exe", &ver[0], &ver[1], &ver[2], &ver[3]);
    get_file_version ("PTD.dll", &ptdver[0], &ptdver[1], 
				 &ptdver[2], &ptdver[3]);
    /* XXX
    if (ptdver[0] != ver[0] || ptdver[1] != ver[1]|| ptdver[2] != ver[2]) {
	log_box (_("WinPT Error"), MB_ERR, 
		 _("The PTD.dll file has a different version than WinPT.exe\n"
		   "Please update the PTD.dll to version %d.%d.%d"),
		   ver[0], ver[1], ver[2]);
	return 0;
    }
    */

    if (gpg_md_selftest ()) {
	msg_box (NULL, _("Cryptographic selftest failed."),
		 _("WinPT Error"), MB_ERR);
	return 0;
    }

    s = gpgme_check_version (MIN_GPGME_VER);
    if (!s || !*s) {
	msg_box (NULL, _("A newer GPGME version is needed; at least "MIN_GPGME_VER),
		 _("WinPT Error"), MB_ERR);
	return 0;
    }

    CreateMutex (NULL, TRUE, PGM_NAME);
    if (GetLastError () == ERROR_ALREADY_EXISTS)
	winpt_inst_found = 1;

    if (cmdline && stristr (cmdline, "--mobile")) {
	msg_box (NULL, "WARNING: mobile modus is not fully implemented yet!", 
		 "WinPT", MB_INFO);
	mobile = 1;
    }

    set_default_kserver ();
    load_gettext (winpt_inst_found);

    if (!mobile) {
	regist_inst_gnupg (1);
	regist_inst_winpt (1, &created);
    }
    else {
	enable_mobile_mode ();
	/* XXX: ask for GPG path */
	created = 1; /* Disable registry writing */
    }

    if (!created) {
	memset (&reg_prefs, 0, sizeof (reg_prefs));
	reg_prefs.use_tmpfiles = 1; /* default */
	reg_prefs.fm.progress = 0; /* XXX: fix the bug and enable it again */
	get_reg_winpt_prefs (&reg_prefs);
	if (!reg_prefs.no_hotkeys)
	    hotkeys_modify ();
	gnupg_load_config ();
    }

    if (is_gpg4win_installed ())
	load_gpg_env (); /* XXX: check return code. */

    rc = gnupg_check_homedir ();
    if (rc) {
	log_box (_("WinPT Error"), MB_ERR, 
	         _("GPG home directory is not set correctly.\n"
		   "Please check the GPG registry settings:\n%s."),
		 winpt_strerror (rc));
	s = get_fileopen_dlg (GetActiveWindow (),
			      _("Select GPG Public Keyring"),
			      _("GPG Keyrings (*.gpg)\0*.gpg\0\0"),
			      NULL);
	if (s != NULL) {
	    size_t n;
	    char *p = strrchr (s, '\\');
	    if (!p)
		BUG (0);
	    n = p - s;
	    if (n) {
		char *file = new char[n+1];
		if (!file)
		    BUG (NULL);
		memset (file, 0, n);
		memcpy (file, s, n);
		file[n] = '\0';		
		set_reg_entry_gpg ("HomeDir", file);
		free_if_alloc (file);
		gnupg_check_homedir (); /* change gpgProgram if needed */
	    }
	}
	else {
	    msg_box (NULL, _("GPG home directory could not be determited."),
		     _("WinPT Error"), MB_ERR);
	    goto start;
	}
    }

    rc = check_gnupg_prog ();
    if (rc) {
	if (msg_box (NULL, _("Could not find the GPG binary (gpg.exe).\n"
			     "Do you want to start the GPG preferences to "
			     "correct  this problem?"), _("WinPT Error"),
			     MB_INFO|MB_YESNO) == IDYES)
	    start_gpgprefs = 1;
	else {
	    msg_box (NULL, winpt_strerror (rc), _("WinPT Error"), MB_ERR);
	    return 0;
	}
    }

    rc = gnupg_access_files ();
    if (!start_gpgprefs && rc) {
	if (rc == WPTERR_GPG_KEYRINGS || rc == WPTERR_GPG_OPT_KEYRINGS) {
	    ec = msg_box (NULL,
		_("Could not access and/or find the public and secret keyring.\n"
		  "If this is an accident, quit the program and fix it.\n\n"
		  "Continue if you want that WinPT offers you more choices.\n"),
		  "WinPT", MB_INFO|MB_YESNO);
	    if (ec == IDYES)
		first_start = 1;
	}
	if (!first_start) {
	    msg_box (NULL, winpt_strerror (rc), _("WinPT Error"), MB_ERR);
	    return 0;
	}
    }

    if (!first_start) {
	rc = gpg_check_permissions (1);
	if (rc && rc == 2)
	    gpg_read_only = 1;
	else if (rc)
	    return 0;
    }
    
    init_gnupg_table ();

    if (fm_parse_command_line (cmdline) > 0) {
	free_gnupg_table ();
	return 0;
    }

    if (cmdline && stristr (cmdline, "--wipe-freespace")) {
	dialog_box_param (glob_hinst, (LPCTSTR)IDD_WINPT_SPACE_SECDEL,
			    GetDesktopWindow(), space_wipefrees_dlg_proc, 0,
			    _("Wipe Free Space"), IDS_WINPT_SPACE_SECDEL);
	free_gnupg_table ();
	return 0;
    }

    load_keyserver_conf (cmdline? 1 : 0);

    if (cmdline && (stristr (cmdline, "--keymanager")
	        || stristr (cmdline, "--cardmanager"))) {
	/* If an instance of WinPT is running, just send the command
	   to open the key manager. Otherwise start a new instance.
	 */
	HWND tray = FindWindow ("WinPT", "WinPT");
	if (stristr (cmdline, "keymanager"))
	    start_manager = ID_WINPT_KEY;
	else
	    start_manager = ID_WINPT_CARD;
	if (tray != NULL) {
	    PostMessage (tray, WM_COMMAND, start_manager, 0);
	    free_gnupg_table ();
	    return 0;
	}
    }

    /* If we found another WinPT instance, just quit to avoid it
       will be executed twice. */
    if (winpt_inst_found) {
	log_debug ("%s", "WinMain: WinPT is already running.");	
	free_gnupg_table ();
	return 0;
    }

    if (cmdline) {
	if (stristr (cmdline, "--enable-debug") || 
	    stristr (cmdline, "--debug")) {
	    gpg_set_debug_mode (1);
	    winpt_debug_msg ();
	    debug = 1;
	}
    }

    wc.hIcon = LoadIcon (glob_hinst, MAKEINTRESOURCE (IDI_WINPT));
    rc = RegisterClass (&wc);
    if (rc == FALSE) {
	msg_box (NULL, _("Could not register window class"), 
		 _("WinPT Error"), MB_ERR);
	free_gnupg_table ();
	return 0;
    }

    hwnd = CreateWindow (PGM_NAME, 
			 PGM_NAME, 
			 0, 0, 0, 0, 0, 
			 NULL, 
			 NULL, 
			 hinst, 
			 NULL);
    if (hwnd == NULL) {
    	msg_box (NULL, _("Could not create window"), _("WinPT Error"), MB_ERR);
	free_gnupg_table ();
	return 0;
    }
    glob_hwnd = hwnd;
    UpdateWindow (hwnd);

    if (!first_start && !start_gpgprefs) {
	gnupg_backup_options ();	
	if (!check_crypto_engine ()) {
	    DestroyWindow (hwnd);
	    free_gnupg_table ();
	    return 0;
	}
    }
    
    if (start_gpgprefs) {
	char *ring;
	DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_GPGPREFS, hwnd,
			gpgprefs_dlg_proc, 0);
	ring = get_gnupg_keyring (0, !NO_STRICT);
	if (gnupg_access_keyring (0) == -1 && get_file_size (ring) == 0)
	    first_start = 1; /* The keyring is empty! */
	free_if_alloc (ring);
    }

    if (first_start) {
	struct first_start_s fs;
	struct genkey_s c;
	HWND h;
start:
	h = GetDesktopWindow ();
	DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_GPGPREFS, h,
			    gpgprefs_dlg_proc, 0);
	DialogBoxParam (glob_hinst, (LPCSTR)IDD_WINPT_FIRST, h,
			first_run_dlg_proc, (LPARAM)&fs);
	switch (fs.choice) {
	case SETUP_KEYGEN:
	    c.interactive = 1;
	    c.first_start = 1;
	    rc = DialogBoxParam (glob_hinst, (LPCSTR)IDD_WINPT_KEYWIZARD,
				 h, keygen_wizard_dlg_proc, (LPARAM)&c);
	    if (!rc)
		goto start;
	    break;

	case SETUP_IMPORT:
	    rc = gnupg_copy_keyrings ();
	    if (rc) {
		msg_box (hwnd, winpt_strerror (rc), _("WinPT Error"), MB_ERR);
		goto start;
	    }
	    break;

	case -1:
	    DestroyWindow (hwnd);
	    free_gnupg_table ();
	    return 0;
	}
	update_keycache (hwnd);
	check_crypto_engine ();
    }
    else {
	gpg_keycache_t c;
	update_keycache (hwnd);
	c = keycache_get_ctx (1);
	if (!c || !gpg_keycache_get_size (c)) {
	    gnupg_display_error ();
	    msg_box (hwnd, _("The keycache was not initialized or is empty.\n"
			     "Please check your GPG config (keyrings, pathes...)"),
			     _("WinPT Error"), MB_ERR);
	    ec = msg_box (NULL, _("It seems that GPG is not set properly.\n"
				  "Do you want to start the GPG preferences dialog?"),
			    "WinPT", MB_INFO|MB_YESNO);
	    if (ec == IDYES) {
		DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_GPGPREFS, hwnd,
				gpgprefs_dlg_proc, 0);
		update_keycache (hwnd);
	    }
	    else {
		DestroyWindow (hwnd);
		free_gnupg_table ();
		return 0;
	    }
	}
	if (check_default_key (c)) {
	    char *p = get_gnupg_default_key ();
	    log_box (_("WinPT Error"), MB_ERR, 
		     _("Default key from the GPG options file could not be found.\n"
		       "Please check your gpg.conf (options) to correct this:\n\n"
		       "%s: public key not found."), p? p : "[null]");
	    free_if_alloc (p);
	    DestroyWindow (hwnd);
	    free_gnupg_table ();
	    return 0;
	}
	if (count_insecure_elgkeys ())
	    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_ELGWARN, glob_hwnd,
			    elgamal_warn_dlg_proc, 0);
    }

    if (start_manager)
	PostMessage (hwnd, WM_COMMAND, start_manager, 0);

    accel_tab = LoadAccelerators (glob_hinst, (LPCTSTR)IDR_WINPT_ACCELERATOR);
    keyring_check_last_access (); /* init */
    while (GetMessage (&msg, hwnd, 0, 0)) {
	if (!TranslateAccelerator (msg.hwnd, accel_tab, &msg)) {
	    TranslateMessage (&msg);
	    DispatchMessage (&msg);
	}
    }
	
    return 0;
}
