/* wptClipVerifyDlg.cpp - WinPT verify dialog
 *	Copyright (C) 2001-2005 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <time.h>

#include "resource.h"
#include "wptGPG.h"
#include "wptTypes.h"
#include "wptCommonCtl.h"
#include "wptKeylist.h"
#include "wptW32API.h"
#include "wptKeyserver.h"
#include "wptNLS.h"
#include "wptContext.h" /* for passwd_s */
#include "wptDlgs.h"
#include "wptErrors.h"
#include "wptVersion.h"

/* XXX use out directly and do not use gpg_data_extract to store the plain text. */


/* Verify data from the clipboard. If @is_detached is set, a detached
   signature is assumed with the data the signature was calculated over
   in @det_data. The context is returned in @r_ctx and the signature
   in @r_sig.
   Return value: 0 on success. */
gpgme_error_t
gpg_clip_verify (int is_detached,
		 const char *det_data, size_t det_len,
		 gpgme_ctx_t *r_ctx,
		 gpgme_signature_t *r_sig)
{
    gpgme_error_t err;
    gpgme_ctx_t ctx;
    gpgme_data_t dat = NULL;
    gpgme_data_t out = NULL;
    gpgme_data_t sig = NULL;
    gpgme_verify_result_t res;

    if (is_detached) {
	err = gpgme_data_new_from_mem (&dat, det_data, det_len, 1);
	if (err)
	    return err;
    }
    
    err = gpgme_new (&ctx);
    if (err)
	goto leave;
    err = gpg_data_new_from_clipboard (&sig, 0);
    if (err)
	goto leave;
    err = gpgme_data_new (&out);
    if (err)
	goto leave;
    
    err = gpgme_op_verify (ctx, sig, dat, out);
    if (err)
	goto leave;
    res = gpgme_op_verify_result (ctx);
    if (!res || !res->signatures) {
	err = gpg_error (GPG_ERR_NO_DATA);
	goto leave;
    }
    *r_sig = res->signatures;
    *r_ctx = ctx;

leave:
    if (err) {
	gpgme_release (ctx);
	*r_ctx = NULL;
    }
    if (dat)
	gpgme_data_release (dat);
    if (out)
	gpgme_data_release (out);
    gpgme_data_release (sig);
    return err;
}


/* Display the policy URL and the notation data of a signature.
   If @not is NULL, it is assumed there is no data. 
   @dlg is the handle to the calling dialog. */
static void
show_notation_data (HWND dlg, gpgme_sig_notation_t nota)
{
    gpgme_sig_notation_t n;
    size_t len=0;
    char *p;

    for (n=nota; n; n = n->next) {
	if (n->name)
	    len += strlen (n->name) + 1 + 2;
	else
	    len += strlen ("policy URL") + 1 + 2;
	len += strlen (n->value) + 1 + 2;
	len += 6;
    }
    p = (char *)calloc (1, len+64);
    if (!p)
	BUG (NULL);
    strcpy (p, "Notation data:\n");
    for (n=nota; n; n = n->next) {
	if (!n->name)
	    strcat (p, "policy url: ");
	else {
	    strcat (p, n->name);
	    strcat (p, " : ");
	}
	strcat (p, "\"");
	strcat (p, n->value);
	strcat (p, "\"\n");
    }

    msg_box (dlg, p, _("Signature Information"), MB_INFO);
    free (p);
}


/* Dialog procedure for the clipboard verification. */
BOOL CALLBACK
clip_verify_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static listview_ctrl_t lv = NULL;
    static text_input_s *ctx = NULL;
    gpgme_error_t err;
    gpgme_signature_t sig = NULL, s;
    gpg_keycache_t kc = NULL;
    gpgme_ctx_t c=NULL;
    const char *det_data=NULL;
    u16 port = HKP_PORT;
    int rc = 0, det_len=0;
    
    switch( msg ) {
    case WM_INITDIALOG:
        SetWindowText (dlg, _("Verify"));
	SetDlgItemText (dlg, IDC_VERIFY_SAVE, _("&Save"));
        kc = keycache_get_ctx (KEYCACHE_PUB);
        if (!kc)
	    BUG (NULL);
        ctx = (text_input_s *)lparam;
	if (ctx) {
	    det_data = ctx->data;
	    det_len = ctx->length;
	}
	err = gpg_clip_verify (ctx && ctx->length > 0, 
			       det_data, det_len, &c, &sig);
        if (err) {
	    msg_box (dlg, gpgme_strerror (err), _("Verify"), MB_ERR);
	    if (c)
		gpgme_release (c);
            EndDialog (dlg, FALSE);
            return FALSE;
        }

	if (gpgme_err_code (sig->status) == GPG_ERR_NO_PUBKEY) {
	    const char *keyid;
	    const char *kserv;
	    if (!sig->fpr)
		BUG (NULL);
	    if (strlen (sig->fpr) == 40)
		keyid = sig->fpr+24;
	    else
		keyid = sig->fpr+16;
	    rc = log_box (_("Verify"), MB_INFO|MB_YESNO,
			  _("Signature made %s using %s key ID 0x%s\n"
			  "Cannot check signature: public key not found\n\n"
			  "Do you want to try to retrieve the key from the keyserver?"),
			  strtimestamp (sig->timestamp),
			  get_key_pubalgo (sig->pubkey_algo), keyid+8);
            if (rc == IDNO) {
                msg_box (dlg, get_gpg_sigstat (GPGME_SIGSUM_KEY_MISSING), _("Verify"), MB_WARN);
		gpgme_release (c);
                EndDialog (dlg, FALSE);
                return FALSE;
            }
	    if (0) {
	    /* FIXME: does GPGME include the keyserver status
	    kserv = gpgme_sig_get_string_attr (sig, GPGME_ATTR_KEYSERVER);
	    if (kserv && strncmp (kserv, "hkp://", 6)) {
		rc = log_box (_("Verify"), MB_INFO|MB_YESNO,
			      _("The users preferred keyserver is '%s'.\n"
			        "Do you want to use it to fetch the key?"), kserv);
		if (rc == IDNO) {
		    kserv = default_keyserver;
		    port = default_keyserver_port;
		}*/
	    }
	    else {
		kserv = default_keyserver;
		port = default_keyserver_port;
	    }
            if (!hkp_recv_key (dlg, kserv, port, keyid+8, 0, 0)) {
		keycache_reload (dlg);
		kc = keycache_get_ctx (KEYCACHE_PUB);
		if (!kc)
		    BUG (dlg);
	    }
        }
	else if (gpgme_err_code (sig->status) == GPG_ERR_BAD_SIGNATURE && 
		 !sig->timestamp)
		 ;
	else if (!sig->timestamp) {
            msg_box (dlg, _("Invalid signature state."), _("Verify"), MB_ERR);
	    gpgme_release (c);
            EndDialog (dlg, FALSE);
            return FALSE;
        }
        verlist_build (&lv, GetDlgItem (dlg, IDC_VERIFY_SIGLIST), 0);

	for (s = sig; s; s = s->next) {
	    rc = verlist_add_sig (lv, s);
	    if (rc)
		msg_box (dlg, _("Could not extract key or signature information."), 
			 _("Verify"), MB_ERR);
	}
	if (sig->exp_timestamp > (DWORD)time (NULL))
	    SetDlgItemText( dlg, IDC_VERIFY_INFO, _("The signature is expired!") );
	if (sig->notations)
	    show_notation_data (dlg, sig->notations);
	gpgme_release (c);
        SetForegroundWindow (dlg);
        return TRUE;
        
    case WM_DESTROY:
        if (lv) {
            listview_release (lv);
            lv = NULL;
        }
        return FALSE;
        
    case WM_SYSCOMMAND:
        if (LOWORD (wparam) == SC_CLOSE)
            EndDialog (dlg, TRUE);
        return FALSE;
        
    case WM_COMMAND:
        switch (LOWORD (wparam )) {
        case IDOK:
            EndDialog (dlg, TRUE);
            return FALSE;
            
        case IDC_VERIFY_SAVE:
            dialog_box_param (glob_hinst, (LPCTSTR)IDD_WINPT_FILE_SAVE, dlg,
                              file_save_dlg_proc, 0, _("Save Plaintext"), 
			      IDS_WINPT_FILE_SAVE);
            break;
        }
        break;
    }
    
    return FALSE;
}

