/* wptSigList.cpp - Listview for key signatures
 *	Copyright (C) 2001-2005 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <time.h>

#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptKeylist.h"
#include "wptNLS.h"
#include "wptUTF8.h"
#include "wptErrors.h"
#include "wptTypes.h"
#include "wptW32API.h"


/* Is the given signature an user-id signature? */
#define IS_UID_CERT(_class) ((_class) >= 0x10 && (_class) <= 0x13)


/* Create a signature list control in @lv with the parent window
   given in @ctrl.
   Return value: 0 on success. */
static int
siglist_build (listview_ctrl_t * lv, HWND ctrl)
{
    struct listview_ctrl_s *c;
    struct listview_column_s implist[] = {
	{0, 240, (char *)_("User ID")},
	{1,  50, (char *)_("Valid")},
	{2,  44, (char *)_("Class")},
	{3,  68, (char *)_("Creation")},
	{4,  80, (char *)_("Key ID")},
	{5,  68, (char *)_("Expiration")},
	{6,  56, (char *)_("Algorithm")},
	{0,   0, NULL}
    };	
    int j, rc = 0;
    
    rc = listview_new (&c);
    if (rc)
	return rc;
    c->ctrl = ctrl;
    for ( j=0; implist[j].fieldname != NULL; j++ )
	listview_add_column( c, &implist[j] );
    listview_set_ext_style( c );
    *lv = c;
    return 0;
}


/* Delete the signature control in @lv. */
void 
siglist_delete( listview_ctrl_t lv )
{
    if (lv) {
	listview_release (lv);
    }
}


/* Indent the string in @old with the level @ilvl.
    Return value: indented string. */
static char *
indent_string (const char * old, int ilvl)
{
    char * p;
    int i;

    p = new char[strlen (old) + 1 + ilvl];
    if (!p)
	BUG (NULL);
    for (i = 0; i < ilvl; i++)
	p[i] = ' ';
    strcpy (p+i, old);
    return p;
}


/* Add an userid signature @ks from the userid @uid to the sig list 
   control @lv.
   Return value: 0 on success. */
static int
siglist_add_key (listview_ctrl_t lv, gpgme_user_id_t uid, gpgme_key_sig_t ks)
{
    gpgme_key_t key=NULL;
    char t[128];
    const char *attr;
    const char *s;
    int rc = 0, key_attr;

    if (ks && !IS_UID_CERT (ks->sig_class))
	return 0;

    rc = listview_add_item (lv, " ");
    if (rc)
	return rc;

    attr = NULL;
    if (!uid) {
	get_pubkey (ks->keyid, &key);
	if (key)
	    attr = key->uids->uid;
    }
    else
	attr = uid->uid;
    if (attr && strlen (attr)) {
	char *uid_utf8 = utf8_to_wincp (attr, strlen (attr));
	char *p = uid? m_strdup (uid_utf8) : indent_string (uid_utf8, 2);
	listview_add_sub_item (lv, 0, 0, p);
	free (uid_utf8);
	free_if_alloc (p);
    }
    else
	listview_add_sub_item (lv, 0, 0, _("  user ID not found"));

    if (uid)
	return 0;

    switch (gpgme_err_code (ks->status)) {
    case GPG_ERR_NO_ERROR:     s = "YES";   break;
    case GPG_ERR_NO_PUBKEY:    s = "NOKEY"; break;
    case GPG_ERR_BAD_SIGNATURE:s = "NO";    break;
    default:                   s = "ERROR"; break;
    }
    listview_add_sub_item( lv, 0, 1, s );

    switch (ks->sig_class) {
    case 0x10: strcpy (t, " "); break;
    case 0x11: strcpy (t, "1"); break;
    case 0x12: strcpy (t, "2"); break;
    case 0x13: strcpy (t, "3"); break;
    }
        
    strcat (t, " ");
    if (!ks->exportable)
	strcat (t, "L");
    /*if (key_attr & GPGME_SIGF_NREV)
	strcat (t, "R");*/
    listview_add_sub_item (lv, 0, 2, t);

    key_attr = ks->timestamp;
    if( key_attr ) {
	s = get_key_created( key_attr );
	listview_add_sub_item( lv, 0, 3, s );
    }
    
    attr = ks->keyid;
    if( attr && strlen( attr ) == 16 ) {
	_snprintf (t, sizeof t -1, "0x%s", attr + 8);
	listview_add_sub_item( lv, 0, 4, t );
    }

    key_attr = ks->expires;
    if (key_attr) {
	s = get_key_created (key_attr);
	listview_add_sub_item (lv, 0, 5, s);
    }

    attr = get_key_pubalgo (ks->pubkey_algo);
    if( attr )
	listview_add_sub_item( lv, 0, 6, (char *)attr );
    
    return 0;
}


/* Load a signature list with the signatures of the key with
   the keyID @keyid. The parent window is in @ctrl.
   Return value is the handle to the listview control. */
listview_ctrl_t
siglist_load (HWND ctrl, const char *keyid)
{    
    gpgme_key_t key;
    gpgme_user_id_t u;
    gpgme_key_sig_t s;
    listview_ctrl_t lv;
    int rc;

    if (siglist_build (&lv, ctrl))
	BUG (NULL);
    rc = get_pubkey(keyid, &key);
    if (rc)
	BUG (NULL);

    /* XXX: the root item is at the end but should come first. */
    for (u=key->uids; u; u = u->next) {
	siglist_add_key (lv, u, NULL);
	for (s = u->signatures; s; s = s->next) {
	    rc = siglist_add_key (lv, NULL, s);
	    if (rc)
		break;
	}
    }
    if (rc) {
	siglist_delete (lv);
	lv = NULL;
    }
    return lv;
}
